from enum import IntEnum


# fmt: off
class Piki(IntEnum):
    NONE            = 0
    RED             = 1
    BLUE            = 2
    YELLOW          = 3
    RED_PLUCKED     = 5
    BLUE_PLUCKED    = 6
    YELLOW_PLUCKED  = 7
# fmt: on
